<?php
/*
 * This file is part of the MyID RapID package.
 */
namespace Intercede;

class Rapid 
{
    /* ******************************************************************************************

        Web server configuration:

        Your we server will need to be configured for Two way TLS.
        For apache configuration see: http://rapidauth.com/docs/reference_apache/
        For ngnix configuration see: http://rapidauth.com/docs/reference_nginx/
        For IIS configuration see: http://rapidauth.com/docs/reference_azure/


        Include RapID SDK:
        
        ----------- using composer -------------
        require 'src/vendor/autoload.php';

        ----------- using Rapid.php and downloading httpful.phar manually -------------
        set_include_path(dirname(__FILE__)); //assuming you place the files in the same directory as your php file.
        require 'rapid.php';
        require 'httpful.phar';


        Usage example:

        $rapid = new \Intercede\Rapid(load_key());
        
        .....
        
        //how to request a credential using an anonymous ID
        $id_for_collection = $rapid->request("anonymised_user_id");
        

        .....

        //how to get the authenticated user anonymous ID from certificate CN field
        $authenticated_user_id = $rapid->authenticated_user();

        NOTE: For developement TLS does not need to be configured for authenticated_user() to work. You can 
        provide a parameter to echo back the anonymous_id you want.RapID distinguishes developer mode using php 
        setting display_errors. 
        i.e. the code below will return "developer_anony_id" in developer mode; and in production with no code change
        will return the anonymous id from CN field in certificate.

        $authenticated_user_id = $rapid->authenticated_user("developer_anony_id");

        
        function load_key() 
        {
            // Generate crt.pem from service authentication pfx 
            // openssl pkcs12 -in <path to service authentication pfx> -out crt.pem -clcerts -nokeys -clcerts
        
            // Generate key.pem from service authentication pfx
            // openssl pkcs12 -in rapid.testing.rp.pfx -out key.pem -nocerts

            return array("cert"=>'/filepath/to/client/crt.pem', "key"=>'/filepath/to/client/key.pem', "passphrase"=>'passphrase-for-your-key');
        }

    ****************************************************************************************** */

    // A Rapid object talks to the RapID Auth service.  RapID uses TLS to validate
    // authenticity of messages and must be provided with an authentication certificate.
    // Service authentication certificates are downloaded from the Customer Portal. 
    public function __construct($key)
    {
        $this->rapid_client_key = $key;
        $this->rapid_request_url = $this->buildRapidUrl("request.rapidauth.com/rapid");
    }


    /// The $subject is an anonymised identifier mapped by your website to a user&device.  It 
    /// will be put verbatim into the Common Name of the certificate issued to the device.
    ///
    /// When the device later connects to your website it will present the client certificate via
    /// mutual TLS.  This can be accessed in PHP using
    /// `$_SERVER['SSL_CLIENT_S_DN_CN']`
    /// See http://httpd.apache.org/docs/2.4/mod/mod_ssl.html#envvars.
    public function request($subject) {
        $payload = array("SubjectName" => $subject);
        return $this->sendRequest(json_encode($payload));
    }

    public function alive() {
         return "hello RapID world";
    }


    public function authenticated_user($annonID = "rapid_dev_anon_id")
    {
        if (isset($_SERVER['SSL_CLIENT_S_DN_CN'])) // check if CN name of cert is set
        {
            return $_SERVER['SSL_CLIENT_S_DN_CN'];  // return NN name of cert 
        }

        if (isset($_SERVER['SSL_CLIENT_NGINX_DN'])) // check if nginx has passed on DN information
        {
            $matches = array();
            preg_match('/CN=(.*)/', $_SERVER['SSL_CLIENT_NGINX_DN'], $matches);
            if (sizeof( $matches) > 1)
            {
                return $matches[1];
            }
            else
            {
                //If there is no match for CN= but there is some value then return that value!
                if ($_SERVER['SSL_CLIENT_NGINX_DN'] != "")
                {
                    return $_SERVER['SSL_CLIENT_NGINX_DN'];  // return NN name of cert
                }
            }
        }


        //in production mode throw an exception if we have not retrieved successfully the CN value from client auth cert
        if ($_SERVER['RAPID_DEV_MODE'] != '1') 
            throw new \Exception("No cerficiate CN value found" );

        //in dev mode if the annon ID is empty we also thrown an exception
        if ($annonID == "")
        {
             throw new \Exception("In development mode, no certificate or annonID found" );
        }

        return $annonID; //if we end up here then in dev mode and the developer has asked use to echo back the default annonID or the value they passed in
    }

    /****** INTERNAL METHODS *****/

    private function buildRapidUrl($url) {
        $parsed = parse_url($url);

        if(!isset($parsed['host']) && !isset($parsed['path'])) 
            throw new \Exception("Invalid RapID URL");
        if(!isset($parsed['host']) && !trim($parsed['path'])) 
            throw new \Exception("Invalid RapID URL");

        $hostPath = "";
        if(isset($parsed['host'])) $hostPath .= $parsed['host'];
        if(isset($parsed['port'])) $hostPath .= ":" . $parsed['port'];
        if(isset($parsed['path'])) $hostPath .= $parsed['path'];

        return "https://$hostPath/1.0/RequestCredentialEx";        
    }
    
    private function sendRequest($payload) {

        $private_key_path = $this->rapid_client_key['key'];
        $passphrase = $this->rapid_client_key['passphrase'];
        $temp_key_path = null;
        $curl_compiled_for_nss = $this->curl_compiled_with_nss();
                
        try {
            
	        if($curl_compiled_for_nss){
                $temp_key_path = $this->intercept_key_for_nss($private_key_path, $passphrase);
                if(!empty($temp_key_path)){
                   $private_key_path = $temp_key_path;
                   $passphrase = null;
                }
            }
    
            $response = \Httpful\Request::post($this->rapid_request_url)
                ->authenticateWithCert(
                    $this->rapid_client_key['cert'],
                    $private_key_path,
                    $passphrase
                )
                ->body($payload)
                ->sendsJson()
                ->expectsJson()
                ->send();

        }
        catch(Exception $e) {
            // if exception, catch, clean and re-throw
            $this->clean_up_nss_intercept($temp_key_path);
            throw $e;
        }

        // clean up after intercept if required
        $this->clean_up_nss_intercept($temp_key_path);

        if(isset($response->body->Identifier) || isset($response->body->Message)) {
            // If a valid json identifier, or message is set 
            // The request was satisfied by the Rapid Service but we have an error or a valid
            // Request Id, so return the json object that's been decoded.
            return $response->body;
        }

        // If we have neither of the above, something worse has gone wrong.
        throw new \Exception("Requesting credential failed with HTTP status " . $response->code);
    }

    private function clean_up_nss_intercept($temp_key_path) {
        if(!empty($temp_key_path) && file_exists($temp_key_path)) {
            unlink($temp_key_path);
        }
    }

    private function curl_compiled_with_nss() {
        $curlinfo = curl_version();
        $curl_compile_version = $curlinfo['ssl_version'];
        // If we don't have a version, can't do anything anyway so fall back to standard behaviour
        if(!empty($curl_compile_version)){

            $curl_ssl_parts = explode("/",$curl_compile_version);
            $tls_version = reset($curl_ssl_parts);

            if($tls_version === false) { return false; }

            if($tls_version === "NSS")
            {
                return true;
            }
        }

        return false;
    }

    private function intercept_key_for_nss($private_key_path, $passphrase){
       
        $fp = fopen($private_key_path,"r");
        $keydata = fread($fp, filesize($private_key_path));
        fclose($fp);
       
        $key = openssl_pkey_get_private($keydata, $passphrase);
        
        $certfolder = $this->rapid_client_key['certfolder'];
        $sslconf = $certfolder . '/openssl.cnf';
		if(file_exists($sslconf)) {	
		    $config_args = array('config' => $sslconf); 
		} else {
			$config_args = null;
		}

        $output = "";
        openssl_pkey_export($key, $output, null, $config_args);
       	
        $tempfolder = $this->rapid_client_key['tempfolder'];
    	$temp_key_path    = tempnam($tempfolder, "cr-");
       
        $temp_key_handle = fopen($temp_key_path, "w");
        fwrite($temp_key_handle, $output);
        fclose($temp_key_handle);

        return $temp_key_path;
    }

    private $key;
}
?>
