<?php 
//*************************************************************************************
//*************************************************************************************
// Utility functions for WordPress Rapid Secure Login
//*************************************************************************************
//*************************************************************************************

//*************************************************************************************
// Find the user associated with the given UUID
//*************************************************************************************
function rpsl_get_user_by_uuid( $uuid ) {
// Given a UUID, returns the user object or null if nothing found 
	global $wpdb;
	$uuid       = sanitize_text_field( $uuid );
	$table_name = $wpdb->prefix . "rpsl_devices";
	
	$user_id    = $wpdb->get_var( $wpdb->prepare(
						"SELECT user_id FROM " . $table_name . " WHERE rpsl_uuid = %s ", $uuid ));
	if ($user_id == 0) { 
		return null ;
	} else {
		return get_userdata($user_id);
	}
}

//*************************************************************************************
// Add a new device to a user account
//*************************************************************************************
function rpsl_add_device( $user_id, $rpsl_uuid, $device_name) {
// Adds a device to the user account 
	global $wpdb;
	
	$device_name = sanitize_text_field($device_name);
	$rpsl_uuid   = sanitize_text_field($rpsl_uuid);
	
	if(!is_int($user_id)) {
		rp_trace("rpsl_add_device: user id not int: $user_id");
		return null;
	}
		
	$table_name = $wpdb->prefix . "rpsl_devices";
	$result     = $wpdb->insert($table_name, array(
													'rpsl_uuid'   =>$rpsl_uuid, 
													'user_id'   =>$user_id, 
													'devicename'=>$device_name, 
													'created'   =>current_time('mysql'),
													'last_used' =>current_time('mysql'),
													'status'    =>"ok",
												));
	return $result;
}
//*************************************************************************************
// Set the 'last used' value for the given device
//*************************************************************************************
function rpsl_set_device_last_used ($uuid) {
	global $wpdb;
	$uuid          = sanitize_text_field( $uuid );
	$table_name    = $wpdb->prefix . "rpsl_devices";

	$set_last_used = $wpdb->prepare(
						"UPDATE " . $table_name . " SET last_used = NOW() WHERE rpsl_uuid = %s ", array($uuid)	);
	$updates = $wpdb->query( $set_last_used );
	return;
}

function rpsl_delete_device($uuid){
	global $wpdb;
	$uuid       = sanitize_text_field( $uuid );
	$table_name = $wpdb->prefix . "rpsl_devices";
		
	$delete_query = $wpdb->prepare(
						"DELETE FROM " . $table_name . " WHERE rpsl_uuid = %s ",
						array($uuid)
					);
	$deletes = $wpdb->query( $delete_query );
}

//*************************************************************************************
// Get the password for the Rapid server authentication key file (unique per site)
//*************************************************************************************
function rpsl_keyfile_password() {
	return NONCE_KEY;
}

//*************************************************************************************
// Get the path to the certs folder and ensure it exists	
//*************************************************************************************
function rpsl_cert_folder() {
	$folder = plugin_dir_path( __FILE__ ) . "certs";
	wp_mkdir_p($folder); 
	return $folder;
}

//*************************************************************************************
// Get the path / url to the temp folder and ensure it exists
//*************************************************************************************
function rpsl_temp_folder() {
	$folder = plugin_dir_path( __FILE__ ) . "temp";
	wp_mkdir_p($folder); 
	return $folder;
}

function rpsl_temp_url() {
	return 	plugin_dir_url( __FILE__ ) . "temp";
}

//*************************************************************************************
// Clean up old files
//*************************************************************************************
function rpsl_delete_old_files($folder, $age_in_hours) {
  $files = glob($folder."/*");
  $now   = time();

  foreach ($files as $file)
    if (is_file($file))
      if ($now - filemtime($file) >= 60 * 60 * $age_in_hours) 
        unlink($file);

}

//*************************************************************************************
// Read / Write the Site code, name, login style etc
//*************************************************************************************
function rpsl_site_code() {
	// This is effectively the 'Platform' code
	if(substr(admin_url(), -9) == "wp-admin/") {
		return "W1"; // Standard location
	} else {
		return "W2"; // Non-standard location
	}
}

//-------------------------------------------------------------------------------------------------
function rpsl_set_site_name($name) {
	global $wpdb;
	$clean_name = sanitize_text_field( $name );
	$clean_name = str_replace(rpsl_delim(), " ", $clean_name); // Cannot have the QR code delimiter in the name
	$clean_name = str_replace("|", " ", $clean_name);  // .. or our Ajax delimiter
	
	
	$table_name = $wpdb->prefix . "rpsl_config";
	$update  = $wpdb->prepare("UPDATE " . $table_name . " SET site_name = %s ", $clean_name);
	$updated = $wpdb->query( $update );
	return;
}

function rpsl_site_name() {
	global $wpdb;
	$table_name = $wpdb->prefix . "rpsl_config";
	$sql = "SELECT site_name FROM $table_name ;";
	$name = $wpdb->get_var( $sql );
	return htmlentities($name);
}

//-------------------------------------------------------------------------------------------------
function rpsl_set_show_login($show_login) {
	global $wpdb;
	$show_login = sanitize_text_field($show_login);
	if($show_login != "Yes" && $show_login != "No" && $show_login != "Click" ) { return false; }
	
	$table_name = $wpdb->prefix . "rpsl_config";
	$update  = $wpdb->prepare("UPDATE " . $table_name . " SET show_login = %s ", $show_login	);
	$updated = $wpdb->query( $update );
	return true;
}

function rpsl_show_login() {
	global $wpdb;
	$table_name = $wpdb->prefix . "rpsl_config";
	$sql = "SELECT show_login FROM $table_name ;";
	$show_login = $wpdb->get_var( $sql );
	return htmlentities($show_login);
}

//*************************************************************************************
// QR Code Construction
//*************************************************************************************
function rpsl_action_site($action) {
	return $action . rpsl_site_code() . rpsl_delim();
}

function rpsl_action_site_name($action) {
	return $action . rpsl_site_code() . rpsl_delim() . rpsl_site_name() . rpsl_delim();
}

//-------------------------------------------------------------------------------------------------
// Generate the QR code data
function rpsl_qr_data($action, $requestid, $username) {
// Structure is:   vAPP#tV#tV#tV ...
// v=version (currently 'a'), A=action, PP=platform
// fieldcodes are:  n=sitename, r=requestid, u=username, 
//					h= http siteurl, s=https siteurl
	$d = rpsl_delim();
	$qr_data  = "a" . $action . rpsl_site_code();
	$qr_data .= $d . "n" . rpsl_site_name();
	$qr_data .= $d . "r" . $requestid;
	if($username != "") {$qr_data .= $d . "u" . $username;}
	$url = rpsl_admin_url();
	if(strtolower(substr($url,0,8)) == "https://") {
		$qr_data .= $d . "s" . substr($url,8);
	} else {
		$qr_data .= $d . "h" . substr($url,7);
	}
	
	if($action == "S")	{
		$qr_data .= $d . "p" . session_id();
	}
	
	return $qr_data;
}


//-------------------------------------------------------------------------------------------------
// Generate a QR code image from the given data and return the base 64 encoded image
//-------------------------------------------------------------------------------------------------
function rpsl_qr_base64($data, $size = 200) {
	ob_start();
	\PHPQRCode\QRcode::png($data, false, 'L', 4, 2);
	$image_string = base64_encode( ob_get_contents() );
	ob_end_clean();	
	return "data:image/png;base64," . $image_string;
}

//-------------------------------------------------------------------------------------------------
function rpsl_append_admin_url() {
	if (rpsl_site_code() == "W1") { 
		return  (rpsl_delim() . substr(admin_url(), 0, -9)); // Standard location so strip off the "wp-admin/"
	} else {
		return  (rpsl_delim() . admin_url());
	}
}

//-------------------------------------------------------------------------------------------------
function rpsl_admin_url() {
	if (rpsl_site_code() == "W1") { 
		$url = substr(admin_url(), 0, -9); // Standard location so strip off the "wp-admin/"
	} else {
		$url = admin_url();
	}
	if(substr($url, -1) == "/") { $url = substr($url, 0, -1); }
	return $url;
}

//-------------------------------------------------------------------------------------------------
// Returns the URL to the image folder
function rpsl_image_folder_url() {
	$upload = wp_upload_dir();
	return $upload['baseurl'];
}

//-------------------------------------------------------------------------------------------------
// Returns the QR code section delimiter 
function rpsl_delim() {
	return "#";
}


//*************************************************************************************
// Given a UUID, generate a RapID credential request
//*************************************************************************************
function rpsl_get_new_rapid_request($rpsl_uuid, &$rpsl_request_message) {
	// Call to the RapID request function to submit the request and get back the requestID
	$certfolder = plugin_dir_path( __FILE__ ) . "certs";
	$certfile = $certfolder . "/crt.pem"; 
	$keyfile  = $certfolder . "/key.pem"; 
	$rpsl_server_credentials = array(	
									"cert" 			=> $certfile,
									"key"  			=> $keyfile,
									"passphrase" 	=> rpsl_keyfile_password(),
									"certfolder"	=> rpsl_cert_folder(),
									"tempfolder"	=> rpsl_temp_folder(),
								);
	
	$rpsl_rapid_requestid = null;

	$rapidservice = new \Intercede\Rapid($rpsl_server_credentials);
	if (is_object($rapidservice)) {
		try {
			$rpsl_rapid_response = $rapidservice->request( $rpsl_uuid );

			// If the identifier is set, then we return that as request id
			if(isset($rpsl_rapid_response->Identifier)) {
				$rpsl_rapid_requestid = $rpsl_rapid_response->Identifier;
			}
			// If message is set, an error has occurred, set up the correct message
			elseif(isset($rpsl_rapid_response->Message)){
				$rpsl_request_message = rpsl_extract_rapid_service_message($rpsl_rapid_response);
			}
			else
			{
				throw new \Exception("Error: request method returned with no exception, but no message or identifier was set");
			}
		} catch (Exception $e) {			
			rpsl_trace("Error: unable to connect to RapID Service - bad RapID site key? \n" . $e->getMessage() );
			$rpsl_request_message = "An error occurred, please contact your system administrator.";
		}
	} else {
		rpsl_trace("Error: Unable to create the RapID object");
		$rpsl_request_message = "An error occurred, please contact your system administrator.";
	}

	return $rpsl_rapid_requestid;
}

//*************************************************************************************
// Given a Rapid Request Response, Return a normalised error string.
// Must prefix any error message with Error: for some of the other calling code.
//*************************************************************************************
function rpsl_extract_rapid_service_message($rpsl_rapid_response)
{
	$number = $rpsl_rapid_response->Number;
	$message = $rpsl_rapid_response->Message;

	rpsl_trace("Error requesting RapID Id from RapID Service: $number : $message");

	switch($number)
	{
		case 1402:
			return "Error: please contact your system administrator, there are no more remaining licences for new credentials.";
		default:
			return "Error: please contact your system administrator with error code: $number";
		break;
	}
}


//*************************************************************************************
// Check the POST, validate the signed data and return status, UUID and payload
//*************************************************************************************
function rpsl_rapid_verify () {
	global $rpsl_trace_certs;
	$data     = file_get_contents("php://input");         //DIAG	echo "Data is: $data";
	$dataJson = json_decode($data);                       //DIAG	print_r($dataJson);
	$pkcs7    = "";				                          //DIAG	echo "Extracted P7 is: $pkcs7";
	$phone    = "";
	if(property_exists($dataJson, "phone")) {$phone = sanitize_text_field($dataJson->phone);}
	if(property_exists($dataJson, "pkcs7")) {$pkcs7 = $dataJson->pkcs7;}
	
	// Get some filenames for the signing cert and contents
	$certfolder = rpsl_cert_folder();
	$tempfolder = rpsl_temp_folder();
	$carootcert = $certfolder . "/Rapid_TrustedCa.cer"; // Generic
	$untrusted = $certfolder . "/untrusted.pem";        // Any other root cert that won't be matched
	if ( file_exists($carootcert) ) {
		$pkcs7file  = tempnam($tempfolder, "pk7");
		$certfile   = tempnam($tempfolder, "sig");
		$contfile   = tempnam($tempfolder, "con");
		
		// Add the S/MIME Header and save the signed data to file for OpenSSL to read
		$handle1    = fopen($pkcs7file, "w");
		$mimeheader = "MIME-Version: 1.0\r\n" . "Content-Type: application/pkcs7-mime; smime-type=signed-data;\r\n name=smime.p7m\r\n" . "Content-Transfer-Encoding: base64\r\n\r\n";
		fwrite($handle1, $mimeheader);
		fwrite($handle1, $pkcs7);
		fclose($handle1);

		// PKCS7 flags 
		// Do not chain verification of signers certificates: 
		// that is don't use the certificates in the signed message as untrusted CAs.
		$flags = PKCS7_NOCHAIN;
		
		// Now verify and parse the data, If the json pkcs7 written to the file is not a pkcs7 data packet
		// the openssl libary will fail to parse and verify it. If it is a valid packet but signed with
		// incorrect certificates openssl will also fail to parse and verify it.
		$result = openssl_pkcs7_verify($pkcs7file, $flags, $certfile, array($carootcert), $untrusted , $contfile);
		$handle3 = fopen($contfile, "r");
		$data = fread($handle3, filesize($contfile));
		fclose($handle3);	
		rpsl_log_openssl_error();
		
		$signcert = file_get_contents($certfile);
		$certinfo = openssl_x509_parse($signcert);
		$uuid     = substr($certinfo["name"],4);   // Strip off the "/"

		rpsl_set_device_last_used ($uuid);
		
		//DIAG 
		$certAsText = print_r($certinfo, true);
		if ($rpsl_trace_certs) rpsl_trace("Certificate: \n $certAsText"); 
		if ($rpsl_trace_certs) rpsl_trace("Raw cert is: \n $signcert"); 
		rpsl_trace("Data: $data");

		
		// Clean up the temporary files
		unlink($certfile);
		unlink($contfile);
		unlink($pkcs7file);
		
		// If all is ok, pass back raw post data in result array for calling method to use as they see fit.
		// otherwise return an error array.
		$result = array (
					'status'    => 0,
					'message'   => 'ok',
					'uuid'      => $uuid,
					'data'      => $data,
					'phone'     => $phone,
						);
	} else {
		$result = array (
					'status'    => 100,
					'message'   => 'The RapID trusted issuing CA certificate is missing. Please check the WordPress RapID Settings',
					'uuid'      => '',
					'data'      => '',
					'phone'     => '',
						);
	}
	return $result;
}

/*************************************************************
* function to create a token file to trigger the ajax polling.
*************************************************************/
function rpsl_create_session_token_file($session, $content){
		
		//Write a temporary file for the polling script.
		$tempfolder = rpsl_temp_folder();
		rpsl_delete_old_files($tempfolder, 1);  // Housekeeping
		
		$file = $tempfolder . "/" . $session;		
		rpsl_trace("File to be written: $file");
		
		$fp = fopen($file, "w");
		fwrite($fp, $content);
		fclose($fp);
}

/*************************************************************
 * Adding Scripts
/*************************************************************/

function rpsl_adding_scripts() {
	wp_register_script('rpsl_configure_wordpress_registration', plugins_url('scripts/rpsl_configure_wordpress_registration.js', __FILE__), array('jquery'),'1.9', true);
}

/***************************************************************
 * Plugin Upgrade Filter, 
 * Prevent clearing of existing directory before upgrade.
/*************************************************************/
function rpsl_preserve_directory($options) {
    if($options['hook_extra']['plugin'] == 'rapid-secure-login/rpsl_main.php') {
        $options["clear_destination"] = false;
        $options["clear_working"] = false;
        $options["abort_if_destination_exists"] = false;
    }

    return $options;
}
add_filter('upgrader_package_options', 'rpsl_preserve_directory', 999);

/*************************************************************
 * rpsl_UUID class
 * The following class generates VALID RFC 4122 COMPLIANT
 * Version 4 Universally Unique IDentifiers (UUID)
 *************************************************************/
class rpsl_UUID
{
	/*********************************************************
	 * Version 4 UUIDs are pseudo-random.
	 *********************************************************/
	public static function v4() {
		return sprintf('%04x%04x-%04x-%04x-%04x-%04x%04x%04x',
		mt_rand(0, 0xffff), mt_rand(0, 0xffff), // 32 bits for "time_low"
		mt_rand(0, 0xffff),                     // 16 bits for "time_mid"
		mt_rand(0, 0x0fff) | 0x4000,			// 16 bits for "time_hi_and_version",
												// Four most significant bits hold version number 4
		mt_rand(0, 0x3fff) | 0x8000,    		// 16 bits, 8 bits for "clk_seq_hi_res",
												// 8 bits for "clk_seq_low",
												// Two most significant bits hold zero and one for variant DCE1.1
		mt_rand(0, 0xffff), mt_rand(0, 0xffff), mt_rand(0, 0xffff)	// 48 bits for "node"
		);
	}
}